(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2014 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`LevelDiagram`", {"Splines`", "AtomicDensityMatrix`"}];

Begin["`Private`"];

If[$VersionNumber < 7., SplitBy[e_, f_] := Split[e, f[#1] == f[#2] &]];

PixelHeight[r_] := Subtract @@ Flatten[Position[r, Except@{{255, 255, 255} ..}, 1, Heads -> False]][[{-1, 1}]] + 1;

AbsoluteHeight[expr_] := PixelHeight@Rasterize[expr, Background -> White, ImageResolution -> 72][[1, 1]];

AbsoluteWidth[expr_] := PixelHeight@Transpose@Rasterize[expr, Background -> White, ImageResolution -> 72][[1, 1]];

AbsoluteSize[expr_] := PixelHeight /@ {Transpose@#, #}&@Rasterize[expr, Background -> White, ImageResolution -> 72][[1, 1]];

(* Move Text to the top layer of a Graphics object *)
TextOnTop[g_Graphics] :=
    Module[{heads = {AffineHalfSpace, AffineSpace, Annulus, Arrow,
        ArrowBox, Ball, BezierCurve, BezierCurveBox, BoundaryMeshRegion,
        BSplineCurve, BSplineCurveBox, BSplineSurface,
        BSplineSurface3DBox, CapsuleShape, Circle, CircleBox, Cone,
        ConeBox, ConicHullRegion, ConicHullRegion3DBox,
        ConicHullRegionBox, Cuboid, CuboidBox, Cylinder, CylinderBox,
        Disk, DiskBox, DiskSegment, Ellipsoid, FilledCurve,
        FilledCurveBox, GeoGraphics, GeometricTransformation,
        GeometricTransformationBox, GraphicsComplex, GraphicsComplexBox,
        GraphicsGroup, GraphicsGroupBox, HalfLine, HalfPlane, HalfSpace,
        Hexahedron, HexahedronBox, Hyperplane, InfiniteLine,
        InfinitePlane, InfinitePlane, JoinedCurve, JoinedCurveBox,
        Legended, Line, LineBox, LineLegend, MeshRegion, Panel, PanelBox,
        Parallelepiped, Parallelogram, Placed, Point, PointBox, Polygon,
        PolygonBox, Prism, PrismBox, Pyramid, PyramidBox, Raster,
        Raster3D, Raster3DBox, RasterBox, Rectangle, RectangleBox,
        RegularPolygon, Rotate, Scale, Simplex, Sphere, SphereBox,
        SphericalShell, StadiumShape, SurfaceGraphics, Tetrahedron,
        TetrahedronBox, Texture, Translate, Triangle, Tube, TubeBox}, glist, opts},
        (* Layer a copy with only text over a copy with no text *)
        glist = g /. {{_Text -> {}}, {Alternatives @@ (Blank /@ heads) -> {}}};
        (* Combine Epilog and Prolog options for the two copies *)
        opts = DeleteCases[Table[opt -> DeleteCases[Options[#, opt][[1, 2]] & /@ glist, {}], {opt, {Prolog, Epilog}}], _ -> {}];
        Show[glist, Sequence@@opts]
    ];

ConnectArc[{p1_, p2_}, o_] := Cases[
    Flatten@First@ParametricPlot[
        Evaluate[SplineFit[{p1, Mean[{p1, p2}] + o {{0, -1}, {1, 0}}.Abs[p2 - p1], p2}, Bezier][x]],
        {x, 0, 2}
    ],
    line_Line :> Arrow @@ line,
    Infinity
];

TakeFraction[expr_, x_] := Take[expr, Round[x Length@expr]];

DiskTriangle[{x0_, y0_}, n_, s_, r_] :=
    Flatten@Table[
        Disk[s {x, Sqrt[3]/2. y} + {x0, y0}, r],
        {y, 0, n - 1}, {x, -(n - y - 1)/2, (n - y - 1)/2}
    ];

PointTriangle[{x0_, y0_}, n_, s_] := Flatten@Table[
    Point[s {x, Sqrt[3]/2. y} + {x0, y0}],
    {y, 0, n - 1},
    {x, -(n - y - 1) / 2, (n - y - 1) / 2}
];

Options[RibbonArrow] = {
    ArrowheadHeight -> .15,
    ArrowheadWidth -> .04,
    TextStyle -> {},
    FrameOptions -> {FrameStyle -> {Opacity[.2], Thin}, FrameMargins -> 1, RoundingRadius -> 5, Background -> GrayLevel[1, .9]}
};

RibbonArrow[ends:{{x1_, y1_}, {x2_, y2_}}, w_, textLabel_:"", OptionsPattern[]] :=
    Module[
        {ah = (y2 - y1) OptionValue@ArrowheadHeight, aw = (y2 - y1) OptionValue@ArrowheadWidth,
            frameStyleOpts = FilterRules[OptionValue@FrameOptions, Options@Framed],
            defaultTextStyle = {Black, Opacity[1]}
        },
        {
            Polygon[{
                {x1 - w / 2, y1},
                {x1 + w / 2, y1},
                {x2 + w / 2, y2 - ah},
                {x2 + w / 2 + aw, y2 - ah},
                {x2, y2},
                {x2 - w / 2 - aw, y2 - ah},
                {x2 - w / 2, y2 - ah},
                {x1 - w / 2, y1}
            }],
            If[textLabel =!= "",
                Text[Framed[Style[textLabel, "InlineMath", Background -> None, defaultTextStyle, OptionValue@TextStyle], frameStyleOpts], Mean@ends]
            ]
        }
];

(*Options[SquiggleArrow] = {Frequency -> 20, Amplitude -> .01, Arrowheads -> 1};*)

(*SquiggleArrow[{p1 : {p1x_, p1y_}, p2 : {p2x_, p2y_}}, opts : OptionsPattern[]] := Module[*)
    (*{   f = OptionValue[Frequency], a = OptionValue[Amplitude], len = Norm[p2 - p1],*)
        (*arr = (Abs[If[Head[#1] =!= List, {0, #1}, #1]]&)[OptionValue[Arrowheads]], q1 = p1, q2 = p2*)
    (*},*)
    (*If[MatchQ[arr, {x_ /; x != 0, 0}], arr = Reverse[arr];{q1, q2} = Reverse[{q1, q2}]];*)
    (*Quiet[*)
        (*GeometricTransformation[*)
            (*Translate[*)
                (*{   GetLines[ParametricPlot[{a Sin[2\[Pi] f t], t}, {t, a arr[[1]]4.2, len - a arr[[2]]4.2}], Infinity],*)
                    (*If[arr[[2]] == 0., {}, Translate[Polygon[a arr[[2]]5.25{{0, 0}, {.265, -1}, {0, -.9}, {-.265, -1}, {0, 0}}], {0, len}]],*)
                    (*If[arr[[1]] == 0., {}, Polygon[a arr[[1]]5.25{{0, 0}, {.265, 1}, {0, .9}, {-.265, 1}, {0, 0}}]]*)
                (*},*)
                (*q1],*)
            (*RotationTransform[{{0, 1}, q2 - q1}, q1]*)
        (*],*)
        (*{RotationMatrix::"degen"}*)
    (*]*)
(*];*)

(*Options[SquiggleArrow] = {Frequency -> 20, Amplitude -> .01};*)

(*SquiggleArrow[{p1 : {p1x_, p1y_}, p2 : {p2x_, p2y_}}, opts : OptionsPattern[]] := Module[*)
    (*{   f = OptionValue[Frequency], a = OptionValue[Amplitude], len = Norm[p2 - p1],*)
        (*q1 = p1, q2 = p2, t1, t2, t3*)
    (*},*)
    (*t1 = 3 a;*)
    (*t2 = 2 a;*)
    (*t3 = len - 2 t1 - 2 t2;*)
    (*Print[{t1,t2,t3}];*)
    (*Quiet[*)
        (*GeometricTransformation[*)
            (*Translate[*)
                (*{   Arrow[*)
                    (*Join @*)
                    (*GetLines[#, Infinity][[1,1]]&/@*)
                        (*ParametricPlot[{#1, t}, #2]&@@@ {*)
                            (*{0, {t, 0, t1}},*)
                            (*{a Exp[(t - t1 - t2) / t2]^2 Sin[2 Pi f (t - t1)], {t, t1, t1 + t2}},*)
                            (*{a Sin[2 Pi f (t - t1)], {t, t1 + t2, t1 + t2 + t3}},*)
                            (*{a Exp[-(t - (t1 + t2 + t3)) / t2]^2 Sin[2 Pi f (t - t1)], {t, t1 + t2 + t3, t1 + 2 t2 + t3}},*)
                            (*{0, {t, t1 + 2 t2 + t3, 2 t1 + 2 t2 + t3}}*)
                        (*}*)
                (*]*)
                (*},*)
                (*q1],*)
            (*RotationTransform[{{0, 1}, q2 - q1}, q1]*)
        (*],*)
        (*{RotationMatrix::"degen"}*)
    (*]*)
(*];*)

Options[SquiggleArrow] = {SquiggleAmplitude -> .03, SquiggleTaper -> {.3, .1}, SquiggleFrequency -> 8};

SquiggleArrow[{p1 : {p1x_, p1y_}, p2 : {p2x_, p2y_}}, opts : OptionsPattern[]] :=
    Module[{f = OptionValue@SquiggleFrequency, a = OptionValue@SquiggleAmplitude, taper = OptionValue@SquiggleTaper,
        len = Norm[p2 - p1], q1 = p1, q2 = p2, tlist, t0, t1, t2, t3, t4, t5, t, segments},
        If[NumericQ@taper, taper = {taper, taper}];
        tlist = {t0, t1, t2, t3, t4, t5} = {0, .1 First@taper, .9 First@taper, len - Total@taper, .9 Last@taper, .1 Last@taper};
        segments = {0, a Exp[2.5 (t - t1 - t2) / t2]^2 Sin[2 Pi f (t - t1)], a Sin[2 Pi f (t - t1)], a Exp[-2.5 (t - (t1 + t2 + t3)) / t4]^2 Sin[2 Pi f (t - t1)], 0};
        Quiet[
            GeometricTransformation[
                Translate[
                    {Arrow[
                        Join @@ (GetLines[#][[1, 1]] &) /@
                            MapThread[ParametricPlot[{#1, t}, {t, #2, #3}] &, {segments, Accumulate@Most@tlist, Accumulate@Rest@tlist}]
                    ]},
                    q1
                ],
                RotationTransform[{{0, 1}, q2 - q1}, q1]
            ],
            {RotationMatrix::"degen"}
        ]
    ];

QN[state_, sym_, sub_, labs_] :=
    Which[
        sub === True, Subscript[sym, StateLabel@state],
        sub === Primes, Derivative[Position[labs, StateLabel@state, {1}, 1, Heads -> False][[1, 1]] - 1][sym],
        True, sym
    ];

If[$VersionNumber < 7., DeleteDuplicates[expr_] := Tally[expr][[All, 1]]];
DeleteAllDuplicates[e_, patt_] := Delete[e, Flatten[Rest@Position[e, #]& /@ DeleteDuplicates@Cases[e, patt, Infinity], 1]];

SetAttributes[PopAppend, HoldFirst];
PopAppend[list_] := (list = RotateLeft[list]; Last[list]);

xMinMax[pos_, length_] := {
    Min@pos[[All, 1]] - length / 2,
    Max@pos[[All, 1]] + length / 2
};

yMinMax[pos_] := {
    Min@pos[[All, 2]],
    Max@pos[[All, 2]]
};

levelOffsetList[sys_, levelOffsetValue_] := Module[
    {offsetVal = Flatten@{levelOffsetValue}},
    Switch[offsetVal,
        {} | {None},
        ConstantArray[0, Length@sys],
        {_?NumericQ..},
        PadRight[{}, Length@sys, offsetVal],
        {_Rule..},
        Total[Boole@StateMatchQ[sys, #1] StateEvaluate[sys, #2]& @@@ offsetVal]
    ]
];

findLevelPositions[sys_, h_, statelabels_, length_, spacing_, parityOffsetValue_, levelOffsetValue_] := Module[
    {pos, xmin, xmax, width, ymin, ymax, height},
(* Find level positions based on M and Energy.
       Use default values for Energy, HyperfineA, and HyperfineB if not specified. *)
    pos = Transpose@{
        (length + spacing) M /@ sys,
        Diagonal@h /. Thread[Energy /@ statelabels -> Range@Length@statelabels - 1] /. {HyperfineA[_] -> .2, HyperfineB[_] -> .05}
    } /. M[_] -> 0;
    (* Shift levels horizontally if Parity is defined for all sublevels
       and ParityOffset is True or a number or is Automatic for a Toy or Zeeman system with more than two state manifolds *)
    If[DefinedQ[sys, Parity] &&
        (NumericQ[parityOffsetValue] || (parityOffsetValue /. Automatic -> Length@statelabels > 2 && AtomicSystemType[sys] != "HyperfineZeeman")),
        width = -Subtract @@ xMinMax[pos, length];
        pos += Thread@{(width + spacing) / 2 parityOffsetValue Parity[sys] /. {True | Automatic -> 1, Even -> -1, Odd -> 1}, 0};
    ];
    pos += Thread@{levelOffsetList[sys, levelOffsetValue], 0};
    {xmin, xmax} = xMinMax[pos, length];
    width = xmax - xmin;
    {ymin, ymax} = yMinMax@pos;
    height = ymax - ymin;
    {pos, xmin, xmax, width, ymin, ymax, height}
];

makeLevels[pos_, levelStyle_, length_] := Module[{levels, levsty},
    levsty = Replace[EnsureList@#, {} -> {{}}] &@ levelStyle;
    levels = (Line[{#, #} + length {{-1 / 2, 0}, {1 / 2, 0}}]) & /@ pos;
    (* Remove levels if requested by LevelStyle *)
    levsty = PadRight[levsty, Length@levels, levsty];
    levels = ReplacePart[levels, Position[levsty, None | False] -> {}];
    levsty = levsty /. None | False -> {};
    Transpose[{levsty, levels}]
];

makeMLabels[sys_, statelabels_, pos_, ymin_, ymax_, height_, mLabelValue_, mLabelPositionValue_, m_, msub_, labelStyleValue_] := Module[
    {mlabopt, mlabelpos, mlabtext, shortPos, dupPos},
    If[MatchQ[mLabelValue, False | None], Return[{}]];
    (* Standardize MLabel option *)
    mlabopt = PadRight[{}, Length[sys], Replace[mLabelValue, All -> True]];
    (* Standardize MLabelPosition list *)
    mlabelpos = PadRight[{}, Length[sys], mLabelPositionValue];
    mlabelpos = If[ListQ@#, #, {#, {0, 0}}] & /@ mlabelpos;
    (* Convert MLabelPosition list to explicit positions, For Automatic, put labels in the top half
       of the diagram above their respective level lines, and put labels in the bottom half below. *)
    mlabelpos = MapThread[
        (Switch[
            #1[[1]],
            Top, {{#2[[1]], ymax}, 1},
            Bottom, {{#2[[1]], ymin}, -1},
            Above, {#2, 1},
            Below, {#2, -1},
            Automatic, {#2, If[#2[[2]] > height / 2, 1, -1]}
        ] + {#1[[2]], 0} &),
        {mlabelpos, pos}
    ];
    (* Convert True, Automatic, Short, "Bare", False, and None to explict labels *)
    mlabtext = MapThread[
        If[!DefinedQ[#2, M],
            "",
            Switch[#1,
                True | Automatic | Short | "Short",
                QN[#2, m, msub, statelabels] == M[#2],
                "Bare",
                M[#2],
                False | None,
                "",
                _,
                #1
            ]
        ] &,
        {mlabopt, sys}
    ];

    (* Delete duplicate labels except for the label with the largest F *)
    dupPos = SortBy[Range@Length@mlabelpos, {mlabelpos[[#]], F[sys[[#]]]}&];
    dupPos = Flatten[Most /@ SplitBy[dupPos, mlabelpos[[#]]&]];
    mlabtext[[dupPos]] = "";

    (* Take m-labels that have the Short option and find groups with the same y position.
       Convert all in each group to short form except the leftmost *)
    shortPos = Flatten@Position[mlabopt mlabtext, Short Except[""] | "Short" Except[""]];
    shortPos = SortBy[shortPos, Reverse@Flatten@mlabelpos[[#]]&];
    shortPos = Flatten[Rest /@ SplitBy[shortPos, Most@Reverse@Flatten@mlabelpos[[#]]&]];
    mlabtext[[shortPos]] = M[sys][[shortPos]];

    (* Make list of Text objects from label text and positions.
       For "Short" style, align the long-form labels so that the m-number is at the center of the level line.
       Vertical position is shifted up slightly to increase symmetry between labels above and below the level line. *)
    Union@If[
        Union[mlabtext] === {""},
        {},
        MapThread[
            Text[Style[#2, "InlineMath", Background -> None, labelStyleValue], First@#1,
                {If[#3 === Short && Head[#2] === Equal, .75, 0], .1 - 1.3 #1[[2]]}
            ] &,
            {mlabelpos, mlabtext, mlabopt}
        ]
    ]
];

line[{{x1_, y1_}, {x2_, y2_}}, s_] = {x1 + (x2 - x1) s, y1 + (y2 - y1) s};

(* Take a list of vertical arrows, and partition it into sublists that do not overlap in the vertical direction. *)
partitionVerticalArrows[partitions_, {}] := partitions;
partitionVerticalArrows[arrows_] := partitionVerticalArrows[{{First@arrows}}, Rest@arrows];
partitionVerticalArrows[partitions_, arrows_] := Module[{interval, position},
(* Find the vertical extent of the arrows in the current partition *)
    interval = Interval @@ ({by@#, ty@#}& /@ Last@partitions);
    (* Shrink the interval slightly so that we don't get overlaps just at an end point. *)
    interval = Interval[{10.^-10, -10.^-10} + First@interval];
    (* Find the first arrow of the remaining arrows that does not overlap *)
    position = Position[arrows, arr_List /; IntervalIntersection[interval, Interval@{by@arr, ty@arr}] == Interval[], {1}, 1];
    If[position == {},
    (* If there is no such arrow, close the partition and begin a new one with the first of the remaining arrows*)
        partitionVerticalArrows[partitions ~Append~ {First@arrows}, Rest@arrows],
    (* else add arrow to the partition and remove from remaining *)
        partitionVerticalArrows[MapAt[Append[#, arrows[[position[[1, 1]]]]]&, partitions, -1], Delete[arrows, position[[1, 1]]]]
    ]
];

(* Endpoints of an arrow *)
pts[arr_] := arr[[-1, 1]];
(* Bottom endpoint, or left endpoint if arrow is horizontal *)
bpts[arr_] := First@SortBy[pts[arr], Reverse];
bx[arr_] := First@bpts[arr];
by[arr_] := Last@bpts[arr];
(* Top endpoint, or right endpoint if arrow is horizontal *)
tpts[arr_] := Last@SortBy[pts[arr], Reverse];
tx[arr_] := First@tpts@arr;
ty[arr_] := Last@tpts@arr;
(* Left endpoint, or bottom endpoint if arrow is vertical *)
lpts[arr_] := First@Sort[pts[arr]];
lx[arr_] := First@lpts[arr];
ly[arr_] := Last@lpts[arr];
(* Right endpoint, or top endpoint if arrow is vertical *)
rpts[arr_] := Last@Sort[pts[arr]];
rx[arr_] := First@rpts@arr;
ry[arr_] := Last@rpts@arr;
(* arc radius of arrow *)
ar[arr_] := If[arr[[-1, 3]] === ConnectArc, arr[[-1, 2]], 0];

slope[arr_] := Slope@pts@arr;

setTopBottom[arr_, shift_] := arr /. arf[{{x1_, y1_}, {x2_, y2_}}, a__] -> arf[{{shift, y1}, {shift, y2}}, a];
setLeft[arr_, shift_] := arr /. arf[{{x1_, y1_}, {x2_, y2_}}, a__] :> arf[{{x1 + (shift - x1) Boole[x1 < x2], y1}, {x2 + (shift - x2) Boole[x2 < x1], y2}}, a];
setRight[arr_, shift_] := arr /. arf[{{x1_, y1_}, {x2_, y2_}}, a__] :> arf[{{x1 + (shift - x1) Boole[x1 > x2], y1}, {x2 + (shift - x2) Boole[x2 > x1], y2}}, a];

(* Move arrow endpoint that falls inside one or more ranges to the left or right of that range *)
clearEndpoint[arr_, ranges_, {p_, setFunc_, minmax_}] := Module[
    {range = Interval @@ Cases[ranges, {min_, max_} /; min < p@arr < max]},
    If[range === Interval[], arr, setFunc[arr, minmax@First@range]]
];

clearLeftRight[arr_, ranges_] := Fold[clearEndpoint[#1, ranges, #2]&, arr, {{lx, setLeft, Max}, {rx, setRight, Min}}];

DivideList[list_, func_] := {Extract[list, #], Delete[list, #]}& @ Position[list, _?func, {1}, Heads -> False];

adjustQuadrant[slantArrows_, horizArrows_, x_, y_, setFunc_, arrowSpacingValue_, levelLengthValue_] := Module[
    {shiftArrows, xpos},
    shiftArrows = Join[{#, {0, -Abs@slope@#}}& /@ slantArrows, {#, {1, -(rx@# - lx@#)}}& /@ horizArrows];
    (* Sort by y, then x *)
    shiftArrows = SortBy[shiftArrows, Through[{y, x}@First@#] &];
    (* Split into runs of arrows with the same y and overlapping x *)
    shiftArrows = Split[shiftArrows, Chop[y@First@#1 - y@First@#2] == 0 && Abs[x@First@#1 - x@First@#2] < arrowSpacingValue &];
    (* Sort each sublist *)
    shiftArrows = SortBy[#, Last][[All, 1]]& /@ shiftArrows;
    (* Shift each arrow horizontally according to its position in its sublist. *)
    xpos = arrowSpacingValue levelLengthValue Range@Length@# + Mean[x /@ #] & /@ shiftArrows;
    shiftArrows = MapThread[setFunc, {Flatten[shiftArrows, 1], Flatten@xpos}];
    (* Re-separate slanted and horizontal arrows *)
    DivideList[shiftArrows, ar@# == 0 && slope@# != 0 &]
];

adjustArrows[arrows_, arrowSpacingValue_, levelLengthValue_] := Module[
    {vertArrows, aboveHorizArrows, belowHorizArrows, xpos, vertxranges, posSlantArrows, negSlantArrows, otherArrows},

    If[!NumericQ@arrowSpacingValue || arrowSpacingValue == 0 || arrows == {}, Return[arrows]];

    (****** Divide list of arrows into vertical and not vertical arrows ******)
    {vertArrows, otherArrows} = DivideList[arrows, Chop[tx@# - bx@#] == 0 &];

    (****** Space vertical arrows ******)
    (* Sort by x position *)
    vertArrows = SortBy[vertArrows, bx];
    (* Split into sublists of arrows closely spaced in the horizontal direction *)
    vertArrows = Split[vertArrows, Abs[bx@#1 - bx@#2] < arrowSpacingValue&];
    (* Sort each sublist in order of decreasing height *)
    vertArrows = SortBy[#, by@# - ty@# &]& /@ vertArrows;
    (* Partition each sublist into sublists of arrows that don't overlap in the vertical direction *)
    vertArrows = partitionVerticalArrows /@ vertArrows;
    (* Horizontally shift each sub-sublist to space them evenly *)
    xpos = (arrowSpacingValue levelLengthValue (Range@Length@# - (Length@# + 1) / 2) + Mean[tx /@ #[[All,1]]])& /@ vertArrows;
    vertArrows = MapThread[setTopBottom, #]& /@ Transpose@{vertArrows, xpos};
    (* Find the occupied x interval range for each group of vertical arrows *)
    vertxranges = {Min@#, Max@#}& /@ Map[tx, vertArrows, {3}];
    vertArrows = Flatten[vertArrows, 2];

    (* Move slanted and horizontal arrow endpoints out of the way of the vertical arrows *)
    otherArrows = clearLeftRight[#, vertxranges]& /@ otherArrows;

    (****** Divide list of arrows into positive slope, negative slope, and above-line and below-line horizontal arrows ******)
    {aboveHorizArrows, otherArrows} = DivideList[otherArrows, ar@# > 0 || (ar@# == 0 && slope@# == 0) &];
    {belowHorizArrows, otherArrows} = DivideList[otherArrows, ar@# < 0 &];
    {posSlantArrows, negSlantArrows} = DivideList[otherArrows, slope@# > 0 &];

    (* Space the bottom points of positive slope slanted arrows, and the left points of above-line horizontal arrows *)
    {posSlantArrows, aboveHorizArrows} = adjustQuadrant[posSlantArrows, aboveHorizArrows, lx, ly, setLeft, arrowSpacingValue, levelLengthValue];
    (* Space the bottom points of negative slope slanted arrows, and the right points of above-line horizontal arrows *)
    {negSlantArrows, aboveHorizArrows} = adjustQuadrant[negSlantArrows, aboveHorizArrows, rx, ry, setRight, arrowSpacingValue, -levelLengthValue];

    {belowHorizArrows, aboveHorizArrows} = DivideList[Join[belowHorizArrows, aboveHorizArrows], ar@# < 0 || (ar@# == 0 && slope@# == 0) &];

    (* Space the top points of positive slope slanted arrows, and the right points of below-line horizontal arrows *)
    {posSlantArrows, belowHorizArrows} = adjustQuadrant[posSlantArrows, belowHorizArrows, rx, ry, setRight, arrowSpacingValue, -levelLengthValue];
    (* Space the top points of negative slope slanted arrows, and the left points of below-line horizontal arrows *)
    {negSlantArrows, belowHorizArrows} = adjustQuadrant[negSlantArrows, belowHorizArrows, lx, ly, setLeft, arrowSpacingValue, levelLengthValue];

    (* Join the lists of arrows back together. *)
    Join[vertArrows, posSlantArrows, negSlantArrows, aboveHorizArrows, belowHorizArrows]
];

Options[setupArrows] = (# -> {}&) /@ {ExtraArrows, ArcRadius, ArrowOffset, ArrowSetback, ArrowStyle, SuppressOffResonantCouplings,
    MaxArcSlope, ArrowFunction, LevelLength};

setupArrows[sys_, h_, pos_, opts : OptionsPattern[]] := Module[
    {mixing, arcrad, arrowoff, arrowsetback, arrows, arrowfunc, thisarrsty, matelem, freqs, x1, y1, x2, y2, arr, arrsty,
        setback, thisarcrad,
        suppressOffResonantCouplings = OptionValue@SuppressOffResonantCouplings,
        levelLength = OptionValue@LevelLength,
        maxArcSlope = OptionValue@MaxArcSlope},

    (* Find the positions of all off-diagonal nonzero elements of the Hamiltonian, and add all of the requested extra arrows. *)
    mixing = DeleteCases[
        Join[
            Position[h, x_ /; ! PossibleZeroQ[x], {2}, Heads -> False],
            List @@@ Union@Flatten[{Reverse[#], #} & /@ Flatten[Tuples /@ Flatten@Map[Flatten@StatePosition[sys, #]&, Flatten@{OptionValue@ExtraArrows}, {2}]]]
        ],
        {x_, x_}
    ];
    If[mixing == {}, Return[{}]];

    (* Standardize option values *)
    arcrad = RotateRight@EnsureList@OptionValue@ArcRadius;
    arrowoff = Replace[OptionValue@ArrowOffset, {None -> {{0, 0}}, x_?NumericQ -> {{x, 0}}, x : {_?NumericQ, _?NumericQ} -> {x}}];
    arrowoff = Replace[arrowoff, {x : {_?NumericQ, _?NumericQ} -> {x, x}, None -> {0, 0}, x_?NumericQ -> {x, 0}}, {1}];
    arrowsetback = Replace[OptionValue@ArrowSetback, {None -> {{0, 0}}, x_?NumericQ -> {{0, x}}, x : {_?NumericQ, _?NumericQ} -> {x}}];
    arrowsetback = Replace[arrowsetback, {x_?NumericQ -> {0, x}, None -> {0, 0}}, {1}];
    arrsty = Replace[EnsureList@OptionValue@ArrowStyle, {} -> {{}}];
    arrowfunc = OptionValue@ArrowFunction /. Automatic -> (Arrow[#1] &);

    (* Loop over mixing positions, sowing arrow specifications. *)
    arrows = Last@Last@Reap[
        (   {{x1, y1}, {x2, y2}} = pos[[#]];
            (* Discard arrow unless levels are ordered correctly *)
            If[(y1 < y2 || (y1 == y2 && x1 < x2)) && (thisarrsty = PopAppend[arrsty]) =!= None,
            (* Find all terms with explicit coupling frequencies.
                           If there is no term with an explicit coupling frequency,
                           use the transition frequency as the coupling frequency.*)
                matelem = ExpandAll@TrigToExp@Extract[h, #];
                freqs = Union@Cases[
                    {matelem},
                    a_ E^(Complex[0 | 0., x_?NumericQ] w_. t) :> {x, a, w},
                    2
                ] /. {} -> {{(y2 - y1), matelem, w}};
                (* Delete terms with negative coupling frequencies, if requested. *)
                (* FIGURE OUT EXACTLY WHAT I'M DOING HERE? IS THIS CORRECT IN ALL CASES? *)
                If[suppressOffResonantCouplings,
                    freqs = DeleteCases[freqs, f_ /; Negative@First@f]
                ];
                freqs = freqs /. {{x_, a_, w_} /; !NumericQ[w] :> {y2 - y1, a}, {x_, a_, w_} :> {x, a}};
                (* Calculate arrow positions, including offsets *)
                arr = (
                    y2 = y1 + #[[1]];
                    {{x1, y1}, {x2, y2}} +
                        levelLength PopAppend[arrowoff] {{Order[x1, x2], Order[y1, y2]}, {Order[x2, x1], Order[y2, y1]}}
                ) & /@ freqs;
                (* Do arrow setback *)
                arr = (
                    setback = levelLength PopAppend[arrowsetback] / Norm[Subtract @@ #];
                    {line[#, setback[[1]]], line[#, 1 - setback[[2]]]}
                )& /@ arr;
                (* Find arc radius for horizontal arrows *)
                thisarcrad = First[arcrad = RotateLeft@arcrad];
                (* Sow arrow with style*)
                thisarrsty = Replace[thisarrsty, List -> Sequence, {1}, Heads -> True];
                Sow[{
                    thisarrsty,
                    If[Abs@Slope@#1 <= maxArcSlope,
                        arf[#1, thisarcrad, ConnectArc],
                        arf[#1, #2, arrowfunc]
                    ]
                }]& @@@ Transpose@{arr, freqs[[All, 2]]}
            ];
        )& /@ mixing
    ]
];

(* Takes a list of points (opos) and a minimum spacing between them, and returns a list of points where each
point is as close to the original position as possible while being spaced at least minSpacing apart *)
spacePoints[opos_, minSpacing_] := Module[{pos, p, startpos, ord, sortpos},
    (* Sort the points *)
    ord = Ordering@opos;
    sortpos = opos[[ord]];
    If[Min@Differences@sortpos >= minSpacing, Return[opos]];
    (* Evenly space the points, then minimize the distance from the original points, while maintaning a minimum separation *)
    (* Then unsort the points *)
    pos = Array[p, Length[opos]];
    startpos = minSpacing (Range@Length@opos - (Length[opos] - 1) / 2 - 1.);
    sortpos[[ord]] = pos /. Last@FindMinimum[
        {   Total[(pos - sortpos)^2],
            And @@ Thread[Most@pos + minSpacing / 2 < Rest@pos - minSpacing / 2]
        },
        Transpose[{pos, startpos}]
    ];
    sortpos
];

makeJFLabels[sys_, pos_, xmin_, xmax_, statelabels_, quantumNumberSymbolsValue_, quantumNumberSubscriptsValue_,
    labelStyleValue_, levelLengthValue_, levelSpacingValue_, minLabelSpacingValue_, fLabelValue_, jLabelValue_, mLabelValue_,
    stateLabelValue_, stateLabelPositionValue_] := Module[
    {j, f, m, jsub, fsub, msub, fstates, flabeltext, jstates, jlabeltext, statelabeltext, labeltext, positions, labelpos, statepos},
    {j, f, m} = quantumNumberSymbolsValue;
    {jsub, fsub, msub} = quantumNumberSubscriptsValue;
    (* A subset of the system with one state representing each hyperfine level *)
    fstates = Union[sys, SameTest -> (StateLabel@#1 === StateLabel@#2 && F@#1 === F@#2 &)];
    (* Find the label text for each hyperfine level *)
    flabeltext = If[AtomicSystemType[sys] == "HyperfineZeeman" && !MatchQ[fLabelValue, None | False],
        Replace[
            fLabelValue,
            (Automatic | All | True) -> (QN[#, f, fsub, statelabels] == F[#] & /@ fstates)
        ],
        {}
    ];
    (* A subset of the system with one state representing the largest F for each J state *)
    jstates = Last /@ Split[SortBy[fstates, {StateLabel@#, J@#, F@#}&], StateLabel@#1 === StateLabel@#2 && J@#1 === J@#2 &];
    (* Find the label text for each j state *)
    jlabeltext = If[AtomicSystemType[sys] != "Toy" && !MatchQ[jLabelValue, None | False],
        Replace[
            jLabelValue,
            (Automatic | All | True) -> (QN[#, j, jsub, statelabels] == J[#] & /@ jstates)
        ],
        Null
    ];
    statelabeltext = If[!MatchQ[stateLabelValue, None | False],
        Replace[
            stateLabelValue,
            (Automatic | All | True) -> (StateLabel@jstates)
        ],
        Null
    ];
    statelabeltext = If[jlabeltext === statelabeltext === Null,
        {},
        Row[#, ","] & /@ Thread[PadRight[{}, Length@jstates, #] & /@ {statelabeltext, jlabeltext}] /. Null -> Sequence[]
    ];
    If[flabeltext == {} && statelabeltext == {}, Return[{}]];
    If[flabeltext == {},
        (* J labels only, set positions for each J label at the average of the positions of all sublevels with the largest value of F for that J *)
        labeltext = statelabeltext;
        positions = Extract[pos, #] & /@ StatePosition[sys, StateLabel == StateLabel[#] && F == F[#]] & /@ jstates,
        (* else set positions for each F label at the average of the positions of all sublevels with that F *)
        positions = Extract[pos, #] & /@ StatePosition[sys, StateLabel == StateLabel[#] && F == F[#]] & /@ fstates;
        If[statelabeltext == {},
            (* F labels only *)
            labeltext = flabeltext,
            (* else F and J labels *)
            labeltext = MapThread[
                Row@Join[
                    If[(statepos = Position[jstates, #1]) != {}, {Extract[statelabeltext, First@statepos], ", "}, {}],
                    {#2}
                ]&,
                {fstates, flabeltext}
            ]
        ]
    ];
    If[stateLabelPositionValue === Automatic,
        labelpos = (side = Sign[#1 - Mean[{xmin, xmax}] - 10^-6]; {{side (levelLengthValue + levelSpacingValue)/2 + #1, #2}, {-side, 0}}) & @@@ Mean /@ positions,
        labelpos = {{xmin - levelSpacingValue, #}, {1, 0}} & /@ spacePoints[Last /@ Mean /@ positions, minLabelSpacingValue];
        (* This moves labels farther away if using the "Short" m-label style. Should do a better job here. Or don't offset left-most m-label for short style? *)
        If[mLabelValue === Short | "Short",
            labeltext = Row[{#, Invisible["M="]}]& /@ labeltext
        ]
    ];
    MapThread[
        Text[TraditionalForm@Style[#2, "InlineMath", Background -> None, labelStyleValue], #1[[1]], #1[[2]]] &,
        {labelpos, labeltext}
    ]
];

diskTrianglePopFunc[pos_, pop_, length_] := TakeFraction[DiskTriangle[pos + {0, .075 length}, 6, .15 length, .07 length], pop];

barPopFunc[pos_, pop_, length_] := Rectangle[pos + {-.2 length, 0}, pos + {.2 length, pop}];

makePopulations[rho_, pos_, populationFunctionValue_, populationStyleValue_, levelLengthValue_] := Module[
    {popsty, popfunc, pops},
    If[rho === {{0}}, Return[{}]];
    popsty = Replace[populationStyleValue, List -> Sequence, {1}, Heads -> True];
    popfunc = populationFunctionValue /. {Automatic | "DiskTriangle" -> diskTrianglePopFunc, "Bar" -> barPopFunc};
    pops = MapThread[popfunc[#1, #2, levelLengthValue]&, {pos, Diagonal@rho}];
    {popsty, pops}
];

Options[LevelDiagram] =
    Join[#, FilterRules[Options[Graphics], Except@#]] &@{
        ArrowStyle -> {}, ArrowSpacing -> .1, ExtraArrows -> {}, LevelStyle -> {},
        LevelLength -> .8, LevelSpacing -> .2, MLabel -> None,
        MLabelPosition -> Automatic, FLabel -> None, JLabel -> None, StateLabel -> None, StateLabelPosition -> Left,
        LabelStyle -> {}, MinLabelSpacing -> 0.3,
        ParityOffset -> Automatic, LevelOffset -> None, MaxArcSlope -> .2,
        ArcRadius -> -.5, SuppressOffResonantCouplings -> True,
        PopulationFunction -> Automatic, PopulationStyle -> {},
        ArrowFunction -> Automatic, ArrowOffset -> {0, 0},
        ArrowSetback -> 0, QuantumNumberSymbols -> {J, F, M},
        QuantumNumberSubscripts -> {True, True, False}, TextOnTop -> True,
        ImageSize -> {{350}, {100}}
    };

LevelDiagram[sys_?AtomicSystemQ, opts : OptionsPattern[]] := LevelDiagram[sys, Energy@sys, {{0}}, opts];

LevelDiagram[sys_?AtomicSystemQ, h : Except[genericOptionsPattern[1]], opts : OptionsPattern[]] :=
    LevelDiagram[sys, h, {{0}}, opts];

LevelDiagram[sys_?AtomicSystemQ, energies_?VectorQ, rho : Except[genericOptionsPattern[1]], opts : OptionsPattern[]] :=
    LevelDiagram[sys, DiagonalMatrix[energies], rho, opts];

LevelDiagram[sys_?AtomicSystemQ, h : Except[genericOptionsPattern[1]], pops : {_?NumericQ..}, opts : OptionsPattern[]] :=
    LevelDiagram[sys, h, DiagonalMatrix[pops], opts];

LevelDiagram[sys_?AtomicSystemQ, h_?OperatorQ, rho_?OperatorQ, opts : OptionsPattern[]] :=
    LevelDiagram[sys, {{h, FilterRules[{opts}, ExtraArrows]}}, rho, opts];

LevelDiagram[sys_?AtomicSystemQ, hlist : {{_?OperatorQ, genericOptionsPattern[]}..}, rho_?OperatorQ, opts : OptionsPattern[]] := Module[
    {statelabels, arrows, jflabels, levels, mlabels, populations, pos, width, xmax, xmin, ymax, ymin, height, defaultarrsty, defaultpopsty},
    subscripts = PadRight[{}, 3, OptionValue@QuantumNumberSubscripts];

    statelabels = Union@StateLabel@sys;

    (*******Find level positions*******)
    {pos, xmin, xmax, width, ymin, ymax, height} =
        findLevelPositions[sys, First@First@hlist, statelabels,
            OptionValue@LevelLength, OptionValue@LevelSpacing, OptionValue@ParityOffset,
            OptionValue@LevelOffset
        ];

    (*******Make the levels*******)
    levels = makeLevels[pos, OptionValue@LevelStyle, OptionValue@LevelLength];

    (*******Make the M labels*******)
    mlabels = makeMLabels[sys, statelabels, pos, ymin, ymax, height,
        OptionValue[MLabel], OptionValue[MLabelPosition],
        OptionValue[QuantumNumberSymbols][[3]], subscripts[[3]], OptionValue@LabelStyle
    ];

    (*******Make the J and F labels*******)
    jflabels = makeJFLabels[sys, pos, xmin, xmax, statelabels,
        OptionValue@QuantumNumberSymbols, subscripts, OptionValue@LabelStyle,
        OptionValue@LevelLength, OptionValue@LevelSpacing,
        OptionValue@MinLabelSpacing, OptionValue@FLabel, OptionValue@JLabel,
        OptionValue@MLabel, OptionValue@StateLabel, OptionValue@StateLabelPosition
    ];

    (*******Make the arrows*******)
    (* Use options for individual Hamiltonian, then options specified for LevelDiagram, except for ExtraArrows, which
    should only come from the individual Hamiltonian *)
    arrows = Join@@(
        setupArrows[sys, First@#, pos,
            FilterRules[{Rest@#, FilterRules[{opts, Options[LevelDiagram]}, Except[ExtraArrows]]}, Options[setupArrows]]]& /@ hlist
    );
    (* Adjust arrow positions to reduce overlaps *)
    arrows = adjustArrows[arrows, OptionValue@ArrowSpacing, OptionValue@LevelLength];
    (* Apply arrow function *)
    arrows = arrows /. arf -> (#3[#1, #2]&);

    (*******Make the population indicators*******)
    populations = makePopulations[rho, pos, OptionValue@PopulationFunction,
        OptionValue@PopulationStyle, OptionValue@LevelLength];

    (*******Default styles*******)
    defaultarrsty = Directive[RGBColor[.9, 0.01, 0.01], Arrowheads[{-.2, .2} OptionValue@LevelLength/width]];
    defaultpopsty = Directive[EdgeForm[Hue[0.61, 0.54, 0.56]], FaceForm[]];

    (*******Make Graphics*******)
    Graphics[
        {levels, mlabels, jflabels, {defaultpopsty, populations}, {defaultarrsty, arrows}},
        DeleteCases[FilterRules[Join[{opts}, Options@LevelDiagram], Options@Graphics], Alternatives @@ Options@Graphics]
    ] // If[OptionValue@TextOnTop, TextOnTop, Identity]
];

End[];

EndPackage[];
