(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2017 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`WignerRotate`", {"ProcessOptions`", "AtomicDensityMatrix`"}];

Begin["`Private`"];

DownValues[WignerD] = Join[{HoldPattern[WignerD[x_, {al_, be_, ga_}, z___]] :> ADMWignerD[x, {al, be, ga}, z]}, DownValues[WignerD]];

ADMWignerD[s1_AtomicState, {al_, be_, ga_}, s2_AtomicState] :=
    Boole[StateLabel@s1 === StateLabel@s2] *
        Switch[
            AtomicSystemType@s1,
            "Toy",
            1,
            "Zeeman" | "HyperfineZeeman",
            KroneckerDelta[F@s1, F@s2] ADMWignerD[{F@s1, M@s1, M@s2}, {al, be, ga}]
        ];

ADMWignerD[sys_?AtomicSystemQ, {al_, be_, ga_}] := DirectSelfProduct[ADMWignerD[#1, {al, be, ga}, #2]&, sys];

ADMWignerD[j_?HalfIntegerQ | {j_?HalfIntegerQ}, {al_, be_, ga_}] := ADMWignerD[ZeemanSystem@j, {al, be, ga}];

ADMWignerD[{j : {_Integer..}}, {al_, be_, ga_}] := ADMWignerD[#, {al, be, ga}]& /@ j;

ADMWignerD[{j : {{{_Integer..}..}..}}, {al_, be_, ga_}] := Map[ADMWignerD[#, {al, be, ga}]&, j, {3}];

ADMWignerD[{j_?HalfIntegerQ, m_?HalfIntegerQ, mp_?HalfIntegerQ}, {al_, (Pi n_. /; EvenQ[n]) | n_ /; n == 0, ga_}] :=
    KroneckerDelta[m, mp] (-1)^(n j) E^(-I m (al + ga));

ADMWignerD[{j_?HalfIntegerQ, m_?HalfIntegerQ, mp_?HalfIntegerQ}, {al_, Pi n_. /; OddQ[n], ga_}] :=
    KroneckerDelta[-m, mp] (-1)^(n j + m) E^(-I m (al - ga));

ADMWignerD[{j_?HalfIntegerQ, m_?HalfIntegerQ, mp_?HalfIntegerQ}, {al_, be_, ga_}] :=
    E^(-I m al) Wignerd[{j, m, mp}, be] E^(-I mp ga);

Wignerd[{j_?HalfIntegerQ, m1_?HalfIntegerQ, m_?HalfIntegerQ}, be_] := Wignerd[{j, m1, m}, be] =
    Sqrt[((j + m1)!(j - m1)!) / ((j + m)!(j - m)!)] Cos[be / 2]^(-m - m1) Sin[be / 2]^(2j + m + m1) (-1)^(j + m1) *
        Sum[
            (-1)^n Binomial[j + m, n] Binomial[j - m, j + m1 - n] Cos[be / 2]^(2n) Sin[be / 2]^(-2n),
            {n, Max[0, m + m1], Min[j + m, j + m1]}
        ];

Options[TransformTensor] = {StandardOperator -> False};

TransformTensor[{a_, {"OneByOne"}}, {u_, {"Operator"}}, OptionsPattern[]] := a;

TransformTensor[{a_, {"Contravariant", ___}}, {u_, {"Operator"}}, OptionsPattern[]] := u\[Wedge]a;

TransformTensor[{a_, {"Covariant", ___}}, {u_, {"Operator"}}, OptionsPattern[]] := a\[Wedge]ECT@u;

TransformTensor[{a_, {"Operator"}}, {u_, {"Operator"}}, OptionsPattern[]] := If[OptionValue@StandardOperator, ECT@u.a.u, u.a.ECT@u];

TransformTensor[
    {a_, {m : "TensorDecompositionMatrix"..., "TensorDecomposition", "Contravariant"}},
    {u_, {m : "TensorDecompositionMatrix"..., "TensorDecomposition", "Operator"}},
    OptionsPattern[]
] := If[OptionValue@StandardOperator, ECT@u\[CenterDot]a, u\[CenterDot]a];

TransformTensor[
    {a_, {m : "TensorDecompositionMatrix"..., "TensorDecomposition", "Covariant"}},
    {u_, {m : "TensorDecompositionMatrix"..., "TensorDecomposition", "Operator"}},
    OptionsPattern[]
] := If[OptionValue@StandardOperator, a\[CenterDot]u, a\[CenterDot]ECT@u];

TransformTensor[
    {a_, {m : "TensorDecompositionMatrix"..., "TensorDecomposition", "Operator"}},
    {u_, {m : "TensorDecompositionMatrix"..., "TensorDecomposition", "Operator"}},
    OptionsPattern[]
] := ECT@u\[CenterDot]a\[CenterDot]u;

TransformTensor[a : Except@TypedTensorPattern[], u : Except@TypedTensorPattern[], opts : OptionsPattern[]] := Module[
    {aType = TensorType@a, uType = TensorType@u, noMessage = True, result},
    result = Check[
        TransformTensor[{a, aType}, {u, uType}, opts],
        noMessage = False
    ];
    result /; noMessage && TensorTypeQ@aType && TensorTypeQ@uType
];

Options[WignerRotate] = {RotationSense -> "Active", StandardOperator -> False};

WignerRotate[params__, {al_, be_, ga_}, opts : OptionsPattern[]] :=
    WignerRotate[params, {-ga, -be, -al}, RotationSense -> "Active"] /;
        MatchQ[OptionValue@RotationSense, Passive | "Passive"];

(WignerRotate[sys_?AtomicSystemQ, a_, {al_, be_, ga_}, opts : OptionsPattern[]] :=
    TransformTensor[a, WignerD[sys, {al, be, ga}], FilterRules@opts]) // ProcessOptions;

(WignerRotate[a : Except@AtomicSystemPattern[], {al_, be_, ga_}, opts : OptionsPattern[]] :=
    TransformTensor[a, WignerD[{Rank@a}, {al, be, ga}], FilterRules@opts]) // ProcessOptions;

End[];

EndPackage[];
